=== modified file 'HACKING'
--- HACKING	2005-10-06 03:48:08 +0000
+++ HACKING	2005-10-06 04:09:55 +0000
@@ -16,7 +16,10 @@
   In general, you can test at either the command-line level or the
   internal API level.  Choose whichever is appropriate: if adding a
   new command, or a new command option, then call through run_bzr().
-  It is not necessary to do both.
+  It is not necessary to do both. Tests that test the command line level
+  are appropriate for checking the UI behaves well - bug fixes and
+  core improvements should be tested closer to the code that is doing the
+  work. Command line level tests should be placed in 'blackbox.py'.
 
 * Before fixing a bug, write a test case so that it does not regress.
 
@@ -31,7 +34,6 @@
 * Module names should always be given fully-qualified,
   i.e. ``bzrlib.hashcache`` not just ``hashcache``.
 
-
 Evolving interfaces
 -------------------
 
@@ -45,7 +47,6 @@
 Instead, they will get an Attribute error as the api is missing, and
 will know to update their code. If in doubt, just ask on #bzr.
 
-
 Documentation
 =============
 

=== modified file 'NEWS'
--- NEWS	2005-10-05 02:47:53 +0000
+++ NEWS	2005-10-05 05:29:27 +0000
@@ -29,11 +29,15 @@
     * 'bzr status' now shows unknown files inside a selected directory.
       Patch from Heikki Paajanen.
 
+
   INTERNALS:
 
     * Test case failures have the exception printed after the log 
       for your viewing pleasure.
 
+    * InventoryEntry is now an abstract base class, use one of the
+      concrete InventoryDirectory etc classes instead.
+
 
   TESTING:
 

=== modified file 'bzrlib/branch.py'
--- bzrlib/branch.py	2005-10-04 11:19:52 +0000
+++ bzrlib/branch.py	2005-10-06 02:17:49 +0000
@@ -34,7 +34,7 @@
                            DivergedBranches, LockError, UnlistableStore,
                            UnlistableBranch, NoSuchFile)
 from bzrlib.textui import show_status
-from bzrlib.revision import Revision, validate_revision_id, is_ancestor
+from bzrlib.revision import Revision, is_ancestor
 from bzrlib.delta import compare_trees
 from bzrlib.tree import EmptyTree, RevisionTree
 from bzrlib.inventory import Inventory
@@ -156,6 +156,7 @@
         """Subclasses that care about caching should override this, and set
         up cached stores located under cache_root.
         """
+        self.cache_root = cache_root
 
 
 class _Branch(Branch):
@@ -246,11 +247,10 @@
                 store = CompressedTextStore(self._transport.clone(relpath))
             else:
                 store = TextStore(self._transport.clone(relpath))
-            if self._transport.should_cache():
-                from meta_store import CachedStore
-                cache_path = os.path.join(self.cache_root, name)
-                os.mkdir(cache_path)
-                store = CachedStore(store, cache_path)
+            #if self._transport.should_cache():
+            #    cache_path = os.path.join(self.cache_root, name)
+            #    os.mkdir(cache_path)
+            #    store = bzrlib.store.CachedStore(store, cache_path)
             return store
         def get_weave(name):
             relpath = self._rel_controlfilename(name)
@@ -784,9 +784,7 @@
         return self.control_weaves.get_weave('ancestry')
 
     def get_ancestry(self, revision_id):
-        """Return a list of revision-ids integrated by a revision.
-        """
-        # strip newlines
+        """Return a list of revision-ids integrated by a revision."""
         if revision_id is None:
             return [None]
         w = self._get_ancestry_weave()
@@ -1200,9 +1198,6 @@
     def add_pending_merge(self, *revision_ids):
         # TODO: Perhaps should check at this point that the
         # history of the revision is actually present?
-        for rev_id in revision_ids:
-            validate_revision_id(rev_id)
-
         p = self.pending_merges()
         updated = False
         for rev_id in revision_ids:

=== modified file 'bzrlib/clone.py'
--- bzrlib/clone.py	2005-10-02 21:51:29 +0000
+++ bzrlib/clone.py	2005-10-04 04:49:21 +0000
@@ -75,6 +75,10 @@
     assert isinstance(to_location, basestring)
     if basis_branch is not None:
         note("basis_branch is not supported for fast weave copy yet.")
+    if not (branch_from.weave_store.listable()
+            and branch_from.revision_store.listable()):
+        return copy_branch_slower(branch_from, to_location, revision,
+                                  basis_branch)
     history = _get_truncated_history(branch_from, revision)
     if not bzrlib.osutils.lexists(to_location):
         os.mkdir(to_location)
@@ -91,7 +95,6 @@
     return branch_to
 
 
-
 def _get_truncated_history(branch_from, revision):
     history = branch_from.revision_history()
     if revision is None:
@@ -104,13 +107,10 @@
     return history[:idx+1]
 
 def _copy_text_weaves(branch_from, branch_to):
-    # TODO: Handle UnlistableStore and fall back to getting a list of 
-    # all file-ids and copying them one by one.
     copy_all(branch_from.weave_store, branch_to.weave_store)
 
 
 def _copy_revision_store(branch_from, branch_to):
-    # TODO: Copy all revisions mentioned in the ancestry of the selected revision
     copy_all(branch_from.revision_store, branch_to.revision_store)
 
 

=== modified file 'bzrlib/commit.py'
--- bzrlib/commit.py	2005-10-02 22:47:02 +0000
+++ bzrlib/commit.py	2005-10-06 02:51:02 +0000
@@ -44,11 +44,6 @@
 # TODO: Update hashcache before and after - or does the WorkingTree
 # look after that?
 
-# This code requires all merge parents to be present in the branch.
-# We could relax this but for the sake of simplicity the constraint is
-# here for now.  It's not totally clear to me how we'd know which file
-# need new text versions if some parents are absent.  -- mbp 20050915
-
 # TODO: Rather than mashing together the ancestry and storing it back,
 # perhaps the weave should have single method which does it all in one
 # go, avoiding a lot of redundant work.
@@ -284,7 +279,7 @@
         """Record the parents of a merge for merge detection."""
         pending_merges = self.branch.pending_merges()
         self.parents = []
-        self.parent_trees = []
+        self.parent_invs = []
         self.present_parents = []
         precursor_id = self.branch.last_revision()
         if precursor_id:
@@ -292,7 +287,7 @@
         self.parents += pending_merges
         for revision in self.parents:
             if self.branch.has_revision(revision):
-                self.parent_trees.append(self.branch.revision_tree(revision))
+                self.parent_invs.append(self.branch.get_inventory(revision))
                 self.present_parents.append(revision)
 
     def _check_parents_present(self):
@@ -346,48 +341,20 @@
                     del self.work_inv[file_id]
             self.branch._write_inventory(self.work_inv)
 
-
-    def _find_entry_parents(self, file_id):
-        """Return the text versions and hashes for all file parents.
-
-        Returned as a map from text version to inventory entry.
-
-        This is a set containing the file versions in all parents
-        revisions containing the file.  If the file is new, the set
-        will be empty."""
-        r = {}
-        for tree in self.parent_trees:
-            if file_id in tree.inventory:
-                ie = tree.inventory[file_id]
-                assert ie.file_id == file_id
-                if ie.revision in r:
-                    assert r[ie.revision] == ie
-                else:
-                    r[ie.revision] = ie
-        return r
-
     def _store_snapshot(self):
         """Pass over inventory and record a snapshot.
 
         Entries get a new revision when they are modified in 
         any way, which includes a merge with a new set of
-        parents that have the same entry. Currently we do not
-        check for that set being ancestors of each other - and
-        we should - only parallel children should count for this
-        test see find_entry_parents to correct this. FIXME <---
-        I.e. if we are merging in revision FOO, and our
-        copy of file id BAR is identical to FOO.BAR, we should
-        generate a new revision of BAR IF and only IF FOO is
-        neither a child of our current tip, nor an ancestor of
-        our tip. The presence of FOO in our store should not 
-        affect this logic UNLESS we are doing a merge of FOO,
-        or a child of FOO.
+        parents that have the same entry. 
         """
         # XXX: Need to think more here about when the user has
         # made a specific decision on a particular value -- c.f.
         # mark-merge.  
         for path, ie in self.new_inv.iter_entries():
-            previous_entries = self._find_entry_parents(ie. file_id)
+            previous_entries = ie.find_previous_heads(
+                self.parent_invs, 
+                self.weave_store.get_weave_or_empty(ie.file_id))
             if ie.revision is None:
                 change = ie.snapshot(self.rev_id, path, previous_entries,
                                      self.work_tree, self.weave_store)

=== modified file 'bzrlib/errors.py'
--- bzrlib/errors.py	2005-10-02 21:51:29 +0000
+++ bzrlib/errors.py	2005-10-05 10:02:53 +0000
@@ -27,7 +27,10 @@
             return self.args[0]
         elif len(self.args) == 2:
             # further explanation or suggestions
-            return '\n  '.join([self.args[0]] + self.args[1])
+            try:
+                return '\n  '.join([self.args[0]] + self.args[1])
+            except TypeError:
+                return "%r" % self
         else:
             return `self.args`
 
@@ -203,6 +206,10 @@
 
 class NoSuchFile(TransportError, IOError):
     """A get() was issued for a file that doesn't exist."""
+
+    def __str__(self):
+        return self.msg
+
     def __init__(self, msg=None, orig_error=None):
         import errno
         TransportError.__init__(self, msg=msg, orig_error=orig_error)

=== modified file 'bzrlib/fetch.py'
--- bzrlib/fetch.py	2005-10-03 05:27:22 +0000
+++ bzrlib/fetch.py	2005-10-05 10:02:53 +0000
@@ -82,7 +82,7 @@
 
     count_copied -- number of revisions copied
 
-    count_texts -- number of file texts copied
+    count_weaves -- number of file weaves copied
     """
     def __init__(self, to_branch, from_branch, last_revision=None, pb=None):
         if to_branch == from_branch:
@@ -96,7 +96,8 @@
         self.failed_revisions = []
         self.count_copied = 0
         self.count_total = 0
-        self.count_texts = 0
+        self.count_weaves = 0
+        self.copied_file_ids = set()
         if pb is None:
             self.pb = bzrlib.ui.ui_factory.progress_bar()
         else:
@@ -191,6 +192,7 @@
         self._copy_inventory(rev_id, inv_xml, parents)
         self._copy_ancestry(rev_id, parents)
         self.to_branch.revision_store.add(StringIO(rev_xml), rev_id)
+        mutter('copied revision %s', rev_id)
 
 
     def _copy_inventory(self, rev_id, inv_xml, parent_ids):
@@ -210,29 +212,26 @@
         # in memory until everything's done?  But this way is nicer
         # if it's interrupted.
         for path, ie in inv.iter_entries():
-            if not ie.has_text():
-                continue
             if ie.revision != rev_id:
                 continue
             mutter('%s {%s} is changed in this revision',
                    path, ie.file_id)
-            self._copy_one_text(rev_id, ie.file_id)
-
-
-    def _copy_one_text(self, rev_id, file_id):
-        """Copy one file text."""
-        mutter('copy text version {%s} of file {%s}',
-               rev_id, file_id)
+            self._copy_one_weave(rev_id, ie.file_id)
+
+
+    def _copy_one_weave(self, rev_id, file_id):
+        """Copy one file weave."""
+        mutter('copy file {%s} modified in {%s}', file_id, rev_id)
+        if file_id in self.copied_file_ids:
+            mutter('file {%s} already copied', file_id)
+            return
         from_weave = self.from_weaves.get_weave(file_id)
-        from_idx = from_weave.lookup(rev_id)
-        from_parents = map(from_weave.idx_to_name, from_weave.parents(from_idx))
-        text_lines = from_weave.get(from_idx)
         to_weave = self.to_weaves.get_weave_or_empty(file_id)
-        to_parents = map(to_weave.lookup, from_parents)
-        # it's ok to add even if the text is already there
-        to_weave.add(rev_id, to_parents, text_lines)
+        to_weave.join(from_weave)
         self.to_weaves.put_weave(file_id, to_weave)
-        self.count_texts += 1
+        self.count_weaves += 1
+        self.copied_file_ids.add(file_id)
+        mutter('copied file {%s}', file_id)
 
 
 fetch = Fetcher

=== modified file 'bzrlib/inventory.py'
--- bzrlib/inventory.py	2005-10-04 06:39:39 +0000
+++ bzrlib/inventory.py	2005-10-06 02:51:02 +0000
@@ -146,6 +146,47 @@
              output_to, reverse=False):
         """Perform a diff between two entries of the same kind."""
 
+    def find_previous_heads(self, previous_inventories, entry_weave):
+        """Return the revisions and entries that directly preceed this.
+
+        Returned as a map from revision to inventory entry.
+
+        This is a map containing the file revisions in all parents
+        for which the file exists, and its revision is not a parent of
+        any other. If the file is new, the set will be empty.
+        """
+        def get_ancestors(weave, entry):
+            return set(map(weave.idx_to_name,
+                           weave.inclusions([weave.lookup(entry.revision)])))
+        heads = {}
+        head_ancestors = {}
+        for inv in previous_inventories:
+            if self.file_id in inv:
+                ie = inv[self.file_id]
+                assert ie.file_id == self.file_id
+                if ie.revision in heads:
+                    assert heads[ie.revision] == ie
+                else:
+                    # may want to add it.
+                    # may already be covered:
+                    already_present = 0 != len(
+                        [head for head in heads 
+                         if ie.revision in head_ancestors[head]])
+                    if already_present:
+                        # an ancestor of a known head.
+                        continue
+                    # definately a head:
+                    ancestors = get_ancestors(entry_weave, ie)
+                    # may knock something else out:
+                    check_heads = list(heads.keys())
+                    for head in check_heads:
+                        if head in ancestors:
+                            # this head is not really a head
+                            heads.pop(head)
+                    head_ancestors[ie.revision] = ancestors
+                    heads[ie.revision] = ie
+        return heads
+
     def get_tar_item(self, root, dp, now, tree):
         """Get a tarfile item and a file stream for its content."""
         item = tarfile.TarInfo(os.path.join(root, dp))
@@ -159,6 +200,10 @@
         """Return true if the object this entry represents has textual data.
 
         Note that textual data includes binary content.
+
+        Also note that all entries get weave files created for them.
+        This attribute is primarily used when upgrading from old trees that
+        did not have the weave index for all inventory entries.
         """
         return False
 
@@ -264,8 +309,8 @@
                    self.name,
                    self.parent_id))
 
-    def snapshot(self, revision, path, previous_entries, work_tree, 
-                 weave_store):
+    def snapshot(self, revision, path, previous_entries,
+                 work_tree, weave_store):
         """Make a snapshot of this entry which may or may not have changed.
         
         This means that all its fields are populated, that it has its
@@ -276,7 +321,7 @@
         if len(previous_entries) == 1:
             # cannot be unchanged unless there is only one parent file rev.
             parent_ie = previous_entries.values()[0]
-            if self._unchanged(path, parent_ie, work_tree):
+            if self._unchanged(parent_ie):
                 mutter("found unchanged entry")
                 self.revision = parent_ie.revision
                 return "unchanged"
@@ -289,25 +334,17 @@
         mutter('new revision for {%s}', self.file_id)
         self.revision = revision
         change = self._get_snapshot_change(previous_entries)
+        self._snapshot_text(previous_entries, work_tree, weave_store)
         return change
 
     def _snapshot_text(self, file_parents, work_tree, weave_store): 
+        """Record the 'text' of this entry, whatever form that takes.
+        
+        This default implementation simply adds an empty text.
+        """
         mutter('storing file {%s} in revision {%s}',
                self.file_id, self.revision)
-        # special case to avoid diffing on renames or 
-        # reparenting
-        if (len(file_parents) == 1
-            and self.text_sha1 == file_parents.values()[0].text_sha1
-            and self.text_size == file_parents.values()[0].text_size):
-            previous_ie = file_parents.values()[0]
-            weave_store.add_identical_text(
-                self.file_id, previous_ie.revision, 
-                self.revision, file_parents)
-        else:
-            new_lines = work_tree.get_file(self.file_id).readlines()
-            self._add_text_to_weave(new_lines, file_parents, weave_store)
-            self.text_sha1 = sha_strings(new_lines)
-            self.text_size = sum(map(len, new_lines))
+        self._add_text_to_weave([], file_parents, weave_store)
 
     def __eq__(self, other):
         if not isinstance(other, InventoryEntry):
@@ -331,7 +368,7 @@
     def __hash__(self):
         raise ValueError('not hashable')
 
-    def _unchanged(self, path, previous_ie, work_tree):
+    def _unchanged(self, previous_ie):
         """Has this entry changed relative to previous_ie.
 
         This method should be overriden in child classes.
@@ -502,18 +539,29 @@
         self.text_sha1 = work_tree.get_file_sha1(self.file_id)
         self.executable = work_tree.is_executable(self.file_id)
 
-    def snapshot_revision(self, revision, previous_entries, work_tree,
-                          weave_store):
-        """See InventoryEntry.snapshot_revision."""
-        change = super(InventoryFile, self).snapshot_revision(revision, 
-            previous_entries, work_tree, weave_store)
-        self._snapshot_text(previous_entries, work_tree, weave_store)
-        return change
-
-    def _unchanged(self, path, previous_ie, work_tree):
+    def _snapshot_text(self, file_parents, work_tree, weave_store): 
+        """See InventoryEntry._snapshot_text."""
+        mutter('storing file {%s} in revision {%s}',
+               self.file_id, self.revision)
+        # special case to avoid diffing on renames or 
+        # reparenting
+        if (len(file_parents) == 1
+            and self.text_sha1 == file_parents.values()[0].text_sha1
+            and self.text_size == file_parents.values()[0].text_size):
+            previous_ie = file_parents.values()[0]
+            weave_store.add_identical_text(
+                self.file_id, previous_ie.revision, 
+                self.revision, file_parents)
+        else:
+            new_lines = work_tree.get_file(self.file_id).readlines()
+            self._add_text_to_weave(new_lines, file_parents, weave_store)
+            self.text_sha1 = sha_strings(new_lines)
+            self.text_size = sum(map(len, new_lines))
+
+
+    def _unchanged(self, previous_ie):
         """See InventoryEntry._unchanged."""
-        compatible = super(InventoryFile, self)._unchanged(path, previous_ie, 
-            work_tree)
+        compatible = super(InventoryFile, self)._unchanged(previous_ie)
         if self.text_sha1 != previous_ie.text_sha1:
             compatible = False
         else:
@@ -597,10 +645,9 @@
         """See InventoryEntry._read_tree_state."""
         self.symlink_target = work_tree.get_symlink_target(self.file_id)
 
-    def _unchanged(self, path, previous_ie, work_tree):
+    def _unchanged(self, previous_ie):
         """See InventoryEntry._unchanged."""
-        compatible = super(InventoryLink, self)._unchanged(path, previous_ie, 
-            work_tree)
+        compatible = super(InventoryLink, self)._unchanged(previous_ie)
         if self.symlink_target != previous_ie.symlink_target:
             compatible = False
         return compatible

=== modified file 'bzrlib/selftest/__init__.py'
--- bzrlib/selftest/__init__.py	2005-10-05 02:47:53 +0000
+++ bzrlib/selftest/__init__.py	2005-10-05 10:02:53 +0000
@@ -380,6 +380,10 @@
                 print >>f, "contents of", name
                 f.close()
 
+    def failUnlessExists(self, path):
+        """Fail unless path, which may be abs or relative, exists."""
+        self.failUnless(os.path.exists(path))
+        
 
 class MetaTestLog(TestCase):
     def test_logging(self):

=== modified file 'bzrlib/selftest/blackbox.py'
--- bzrlib/selftest/blackbox.py	2005-10-05 01:46:57 +0000
+++ bzrlib/selftest/blackbox.py	2005-10-05 05:29:27 +0000
@@ -19,7 +19,7 @@
 """Black-box tests for bzr.
 
 These check that it behaves properly when it's invoked through the regular
-command-line interface.  This doesn't actually run a new interpreter but 
+command-line interface. This doesn't actually run a new interpreter but 
 rather starts again from the run_bzr function.
 """
 
@@ -33,6 +33,7 @@
 from bzrlib.selftest import TestCaseInTempDir, BzrTestBase
 from bzrlib.branch import Branch
 from bzrlib.osutils import has_symlinks
+from bzrlib.selftest.HTTPTestUtil import TestCaseWithWebserver
 
 
 class ExternalBase(TestCaseInTempDir):
@@ -676,4 +677,16 @@
             chdir("..")
         else:
             progress("skipping symlink tests")
-            
+
+
+class HttpTests(TestCaseWithWebserver):
+    """Test bzr ui commands against remote branches."""
+
+    def test_branch(self):
+        os.mkdir('from')
+        branch = Branch.initialize('from')
+        branch.commit('empty commit for nonsense', allow_pointless=True)
+        url = self.get_remote_url('from')
+        self.run_bzr('branch', url, 'to')
+        branch = Branch.open('to')
+        self.assertEqual(1, len(branch.revision_history()))

=== modified file 'bzrlib/selftest/test_upgrade.py'
--- bzrlib/selftest/test_upgrade.py	2005-10-04 11:51:06 +0000
+++ bzrlib/selftest/test_upgrade.py	2005-10-05 10:02:53 +0000
@@ -29,12 +29,14 @@
 from bzrlib.upgrade import upgrade
 from bzrlib.selftest.treeshape import build_tree_contents
 
+
 class TestUpgrade(TestCaseInTempDir):
+    
     def test_build_tree(self):
         """Test tree-building test helper"""
         build_tree_contents(_upgrade1_template)
-        self.assertTrue(os.path.exists('foo'))
-        self.assertTrue(os.path.exists('.bzr/README'))
+        self.failUnlessExists('foo')
+        self.failUnlessExists('.bzr/README')
 
     def test_upgrade_simple(self):
         """Upgrade simple v0.0.4 format to v5"""
@@ -72,6 +74,11 @@
         eq(len(rev.parent_ids), 2)
         eq(rev.parent_ids[1], 'wibble@wobble-2')
 
+    def test_upgrade_makes_dir_weaves(self):
+        build_tree_contents(_upgrade_dir_template)
+        upgrade('.')
+        self.failUnlessExists(
+            '.bzr/weaves/dir-20051005095101-da1441ea3fa6917a.weave')
 
 
 _upgrade1_template = \
@@ -180,3 +187,54 @@
         '\xd0\x9a`\x0c*W\x1aR\xc1\x94du\x08(\t\xb0\x91\xdeZ\xa3\x9cU\x9cm\x7f\x8dr\x1d\x10Ot\xb8\xc6\xcf\xa7\x907|\xfb-\xb1\xbd\xd3\xfb\xd5\x07\xeeD\xee\x08*\x02\x00\x00'
     ),
 ]
+
+_upgrade_dir_template = [
+    ( './.bzr/', ),
+    ( './.bzr/README',
+        'This is a Bazaar-NG control directory.\n'
+        'Do not change any files in this directory.\n'
+    ),
+    ( './.bzr/branch-format',
+        'Bazaar-NG branch, format 0.0.4\n'
+    ),
+    ( './.bzr/branch-lock',
+        ''
+    ),
+    ( './.bzr/branch-name',
+        ''
+    ),
+    ( './.bzr/inventory',
+        '<inventory>\n'
+        '<entry file_id="dir-20051005095101-da1441ea3fa6917a" kind="directory" name="dir" />\n'
+        '</inventory>\n'
+    ),
+    ( './.bzr/merged-patches',
+        ''
+    ),
+    ( './.bzr/pending-merged-patches',
+        ''
+    ),
+    ( './.bzr/pending-merges',
+        ''
+    ),
+    ( './.bzr/revision-history',
+        'robertc@robertcollins.net-20051005095108-6065fbd8e7d8617e\n'
+    ),
+    ( './.bzr/stat-cache',
+        '### bzr hashcache v5\n'
+    ),
+    ( './.bzr/text-store/', ),
+    ( './.bzr/inventory-store/', ),
+    ( './.bzr/inventory-store/robertc@robertcollins.net-20051005095108-6065fbd8e7d8617e.gz',
+        '\x1f\x8b\x08\x00\x0c\xa2CC\x02\xff\xb3\xc9\xcc+K\xcd+\xc9/\xaa\xb4\xe3\xb2\x012\x8a*\x15\xd22sR\xe33Sl\x95R2\x8bt\x8d\x0c\x0cL\r'
+        "\x81\xd8\xc0\x12H\x19\xea\xa6$\x1a\x9a\x98\x18\xa6&\x1a\xa7%\x9aY\x1a\x9a'*)dg\xe6A\x94\xa6&\x83LQR\xc8K\xccM\x05\x0b()\xe8\x03\xcd\xd4G\xb2\x00\x00\xc2<\x94\xb1m\x00\x00\x00"
+    ),
+    ( './.bzr/revision-store/', ),
+    ( './.bzr/revision-store/robertc@robertcollins.net-20051005095108-6065fbd8e7d8617e.gz',
+        '\x1f\x8b\x08\x00\x0c\xa2CC\x02\xff\xa5OKj\xc30\x14\xdc\xfb\x14B\x8b\xec\x92<I\xd6\xc7\xc42\x85\xde\xa0\x17(\xb6\xf4\x9c\n'
+        'l\xa9H"\x90\x9c\xbe\xa6\xa9\xa1\x9b\xae\xbax\x0c\xcc\xe71\xd3g\xbc\x85\x12R$.\xadk\xa8\x15\xb3\xa5oi\xc2\\\xc9kZ\x96\x10\x0b9,\xf5\x92\xbf)\xf7\xf2\x83O\xe5\x14\xb1\x1e\xae\xf5BI\x887\x8c5\xe5\xfb{\xf0\x96\xfei>r\x00\xc9\xb6\x83n\x03sT\xa0\xe4<y\x83\xda\x1b\xc54\xfe~T>Ff\xe9\xcc:\xdd\x8e\xa6E\xc7@\xa2\x82I\xaaNL\xbas\\313)\x00\xb9\xe6\xe0(\xd9\x87\xfc\xb7A\r'
+        "+\x96:\xae\x9f\x962\xc6\x8d\x04i\x949\x01\x97R\xb7\x1d\x17O\xc3#E\xb4T(\x00\xa0C\xd3o\x892^q\x18\xbd'>\xe4\xfe\xbc\x13M\x7f\xde{\r"
+        '\xcd\x17\x85\xea\xba\x03l\x01\x00\x00'
+    ),
+    ( './dir/', ),
+]

=== modified file 'bzrlib/selftest/testfetch.py'
--- bzrlib/selftest/testfetch.py	2005-10-02 23:48:11 +0000
+++ bzrlib/selftest/testfetch.py	2005-10-05 06:44:02 +0000
@@ -24,6 +24,7 @@
 from bzrlib.fetch import greedy_fetch
 
 from bzrlib.selftest import TestCaseInTempDir
+from bzrlib.selftest.HTTPTestUtil import TestCaseWithWebserver
 
 
 def has_revision(branch, revision_id):
@@ -98,7 +99,40 @@
 class TestFetch(TestCaseInTempDir):
 
     def test_fetch(self):
-        
         #highest indices a: 5, b: 7
         br_a, br_b = make_branches()
         fetch_steps(self, br_a, br_b, br_a)
+
+
+class TestHttpFetch(TestCaseWithWebserver):
+
+    def setUp(self):
+        super(TestHttpFetch, self).setUp()
+        self.weblogs = []
+
+    def test_fetch(self):
+        #highest indices a: 5, b: 7
+        br_a, br_b = make_branches()
+        br_rem_a = Branch.open(self.get_remote_url(br_a._transport.base))
+        fetch_steps(self, br_rem_a, br_b, br_a)
+
+    def log(self, *args):
+        """Capture web server log messages for introspection."""
+        super(TestHttpFetch, self).log(*args)
+        if args[0].startswith("webserver"):
+            self.weblogs.append(args[0])
+
+    def test_weaves_are_retrieved_once(self):
+        self.build_tree(("source/", "source/file", "target/"))
+        branch = Branch.initialize("source")
+        branch.add(["file"], ["id"])
+        branch.commit("added file")
+        print >>open("source/file", 'w'), "blah"
+        branch.commit("changed file")
+        target = Branch.initialize("target/")
+        source = Branch.open(self.get_remote_url("source/"))
+        source.weave_store.enable_cache = False
+        self.assertEqual(greedy_fetch(target, source), (2, []))
+        weave_suffix = 'weaves/id.weave HTTP/1.1" 200 -'
+        self.assertEqual(1,
+            len([log for log in self.weblogs if log.endswith(weave_suffix)]))

=== modified file 'bzrlib/selftest/testinv.py'
--- bzrlib/selftest/testinv.py	2005-10-03 15:19:25 +0000
+++ bzrlib/selftest/testinv.py	2005-10-06 02:51:02 +0000
@@ -18,10 +18,12 @@
 import os
 
 from bzrlib.branch import Branch
+from bzrlib.clone import copy_branch
+import bzrlib.errors as errors
 from bzrlib.diff import internal_diff
 from bzrlib.inventory import Inventory, ROOT_ID
 import bzrlib.inventory as inventory
-from bzrlib.osutils import has_symlinks
+from bzrlib.osutils import has_symlinks, rename
 from bzrlib.selftest import TestCase, TestCaseInTempDir
 
 
@@ -65,7 +67,7 @@
         ## XXX
 
 
-class TestInventoryEntry(TestCaseInTempDir):
+class TestInventoryEntry(TestCase):
 
     def test_file_kind_character(self):
         file = inventory.InventoryFile('123', 'hello.c', ROOT_ID)
@@ -219,3 +221,138 @@
                           output)
         self.assertEqual(output.getvalue(),
                          "=== target changed 'target1' => 'target2'\n")
+
+
+class TestSnapshot(TestCaseInTempDir):
+
+    def setUp(self):
+        # for full testing we'll need a branch
+        # with a subdir to test parent changes.
+        # and a file, link and dir under that.
+        # but right now I only need one attribute
+        # to change, and then test merge patterns
+        # with fake parent entries.
+        super(TestSnapshot, self).setUp()
+        self.branch = Branch.initialize('.')
+        self.build_tree(['subdir/', 'subdir/file'])
+        self.branch.add(['subdir', 'subdir/file'], ['dirid', 'fileid'])
+        if has_symlinks():
+            pass
+        self.branch.commit('message_1', rev_id = '1')
+        self.tree_1 = self.branch.revision_tree('1')
+        self.inv_1 = self.branch.get_inventory('1')
+        self.file_1 = self.inv_1['fileid']
+        self.work_tree = self.branch.working_tree()
+        self.file_active = self.work_tree.inventory['fileid']
+
+    def test_snapshot_new_revision(self):
+        # This tests that a simple commit with no parents makes a new
+        # revision value in the inventory entry
+        self.file_active.snapshot('2', 'subdir/file', {}, self.work_tree, 
+                                  self.branch.weave_store)
+        # expected outcome - file_1 has a revision id of '2', and we can get
+        # its text of 'file contents' out of the weave.
+        self.assertEqual(self.file_1.revision, '1')
+        self.assertEqual(self.file_active.revision, '2')
+        # this should be a separate test probably, but lets check it once..
+        lines = self.branch.weave_store.get_lines('fileid','2')
+        self.assertEqual(lines, ['contents of subdir/file\n'])
+
+    def test_snapshot_unchanged(self):
+        #This tests that a simple commit does not make a new entry for
+        # an unchanged inventory entry
+        self.file_active.snapshot('2', 'subdir/file', {'1':self.file_1},
+                                  self.work_tree, self.branch.weave_store)
+        self.assertEqual(self.file_1.revision, '1')
+        self.assertEqual(self.file_active.revision, '1')
+        self.assertRaises(errors.WeaveError,
+                          self.branch.weave_store.get_lines, 'fileid', '2')
+
+    def test_snapshot_merge_identical_different_revid(self):
+        # This tests that a commit with two identical parents, one of which has
+        # a different revision id, results in a new revision id in the entry.
+        # 1->other, commit a merge of other against 1, results in 2.
+        other_ie = inventory.InventoryFile('fileid', 'newname', self.file_1.parent_id)
+        other_ie = inventory.InventoryFile('fileid', 'file', self.file_1.parent_id)
+        other_ie.revision = '1'
+        other_ie.text_sha1 = self.file_1.text_sha1
+        other_ie.text_size = self.file_1.text_size
+        self.assertEqual(self.file_1, other_ie)
+        other_ie.revision = 'other'
+        self.assertNotEqual(self.file_1, other_ie)
+        self.branch.weave_store.add_identical_text('fileid', '1', 'other', ['1'])
+        self.file_active.snapshot('2', 'subdir/file', 
+                                  {'1':self.file_1, 'other':other_ie},
+                                  self.work_tree, self.branch.weave_store)
+        self.assertEqual(self.file_active.revision, '2')
+
+    def test_snapshot_changed(self):
+        # This tests that a commit with one different parent results in a new
+        # revision id in the entry.
+        self.file_active.name='newname'
+        rename('subdir/file', 'subdir/newname')
+        self.file_active.snapshot('2', 'subdir/newname', {'1':self.file_1}, 
+                                  self.work_tree, 
+                                  self.branch.weave_store)
+        # expected outcome - file_1 has a revision id of '2'
+        self.assertEqual(self.file_active.revision, '2')
+
+
+class TestPreviousHeads(TestCaseInTempDir):
+
+    def setUp(self):
+        # we want several inventories, that respectively
+        # give use the following scenarios:
+        # A) fileid not in any inventory (A),
+        # B) fileid present in one inventory (B) and (A,B)
+        # C) fileid present in two inventories, and they
+        #   are not mutual descendents (B, C)
+        # D) fileid present in two inventories and one is
+        #   a descendent of the other. (B, D)
+        super(TestPreviousHeads, self).setUp()
+        self.build_tree(['file'])
+        self.branch = Branch.initialize('.')
+        self.branch.commit('new branch', allow_pointless=True, rev_id='A')
+        self.inv_A = self.branch.get_inventory('A')
+        self.branch.add(['file'], ['fileid'])
+        self.branch.commit('add file', rev_id='B')
+        self.inv_B = self.branch.get_inventory('B')
+        self.branch.put_controlfile('revision-history', 'A\n')
+        self.assertEqual(self.branch.revision_history(), ['A'])
+        self.branch.commit('another add of file', rev_id='C')
+        self.inv_C = self.branch.get_inventory('C')
+        self.branch.add_pending_merge('B')
+        self.branch.commit('merge in B', rev_id='D')
+        self.inv_D = self.branch.get_inventory('D')
+        self.file_active = self.branch.working_tree().inventory['fileid']
+        self.weave = self.branch.weave_store.get_weave('fileid')
+        
+    def get_previous_heads(self, inventories):
+        return self.file_active.find_previous_heads(inventories, self.weave)
+        
+    def test_fileid_in_no_inventory(self):
+        self.assertEqual({}, self.get_previous_heads([self.inv_A]))
+
+    def test_fileid_in_one_inventory(self):
+        self.assertEqual({'B':self.inv_B['fileid']},
+                         self.get_previous_heads([self.inv_B]))
+        self.assertEqual({'B':self.inv_B['fileid']},
+                         self.get_previous_heads([self.inv_A, self.inv_B]))
+        self.assertEqual({'B':self.inv_B['fileid']},
+                         self.get_previous_heads([self.inv_B, self.inv_A]))
+
+    def test_fileid_in_two_inventories_gives_both_entries(self):
+        self.assertEqual({'B':self.inv_B['fileid'],
+                          'C':self.inv_C['fileid']},
+                          self.get_previous_heads([self.inv_B, self.inv_C]))
+        self.assertEqual({'B':self.inv_B['fileid'],
+                          'C':self.inv_C['fileid']},
+                          self.get_previous_heads([self.inv_C, self.inv_B]))
+
+    def test_fileid_in_two_inventories_already_merged_gives_head(self):
+        self.assertEqual({'D':self.inv_D['fileid']},
+                         self.get_previous_heads([self.inv_B, self.inv_D]))
+        self.assertEqual({'D':self.inv_D['fileid']},
+                         self.get_previous_heads([self.inv_D, self.inv_B]))
+
+    # TODO: test two inventories with the same file revision 

=== modified file 'bzrlib/selftest/teststore.py'
--- bzrlib/selftest/teststore.py	2005-10-02 21:51:29 +0000
+++ bzrlib/selftest/teststore.py	2005-10-05 06:44:02 +0000
@@ -119,7 +119,6 @@
         store = get_compressed_store()
         test_ignore_get(self, store)
 
-
     def test_total_size(self):
         store = get_compressed_store('.')
         store.add(StringIO('goodbye'), '123123')
@@ -138,6 +137,7 @@
         self.assertEqual(store_a['1'].read(), 'foo')
         self.assertEqual(store_b['1'].read(), 'foo')
 
+
 class TestMemoryStore(TestCase):
     
     def get_store(self):
@@ -176,6 +176,7 @@
         # store_c = RemoteStore('http://example.com/')
         # self.assertRaises(UnlistableStore, copy_all, store_c, store_b)
 
+
 class TestTextStore(TestCaseInTempDir):
     def test_multiple_add(self):
         """Multiple add with same ID should raise a BzrError"""
@@ -190,7 +191,6 @@
         store = get_text_store()
         test_ignore_get(self, store)
 
-
     def test_copy_all(self):
         """Test copying"""
         os.mkdir('a')

=== modified file 'bzrlib/store/__init__.py'
--- bzrlib/store/__init__.py	2005-10-02 21:51:29 +0000
+++ bzrlib/store/__init__.py	2005-10-04 04:37:29 +0000
@@ -86,6 +86,10 @@
             else:
                 yield False
 
+    def listable(self):
+        """Return True if this store is able to be listed."""
+        return hasattr(self, "__iter__")
+
     def get(self, fileids, permit_failure=False, pb=None):
         """Return a set of files, one for each requested entry.
         
@@ -200,6 +204,10 @@
 
     __str__ = __repr__
 
+    def listable(self):
+        """Return True if this store is able to be listed."""
+        return self._transport.listable()
+
 
 class ImmutableMemoryStore(Store):
     """A memory only store."""
@@ -294,11 +302,8 @@
 def copy_all(store_from, store_to):
     """Copy all ids from one store to another."""
     # TODO: Optional progress indicator
-    if not hasattr(store_from, "__iter__"):
-        raise UnlistableStore(store_from)
-    try:
-        ids = [f for f in store_from]
-    except (NotImplementedError, TransportNotPossible):
-        raise UnlistableStore(store_from)
+    if not store_from.listable():
+        raise UnlistableStore(store_from)
+    ids = [f for f in store_from]
     store_to.copy_multi(store_from, ids)
 

=== modified file 'bzrlib/store/weave.py'
--- bzrlib/store/weave.py	2005-10-02 21:51:29 +0000
+++ bzrlib/store/weave.py	2005-10-05 06:44:02 +0000
@@ -32,8 +32,6 @@
 from bzrlib.trace import mutter
 
 
-
-
 class WeaveStore(Store):
     """Collection of several weave files in a directory.
 
@@ -44,7 +42,7 @@
     def __init__(self, transport):
         self._transport = transport
         self._cache = {}
-	self.enable_cache = False
+        self.enable_cache = False
 
 
     def filename(self, file_id):
@@ -68,10 +66,10 @@
     def _put(self, file_id, f):
         return self._transport.put(self.filename(file_id), f)
 
-
     def get_weave(self, file_id):
         if self.enable_cache:
             if file_id in self._cache:
+                mutter("cache hit in %s for %s", self, file_id)
                 return self._cache[file_id]
         w = read_weave(self._get(file_id))
         if self.enable_cache:

=== modified file 'bzrlib/transport/__init__.py'
--- bzrlib/transport/__init__.py	2005-09-29 21:13:03 +0000
+++ bzrlib/transport/__init__.py	2005-10-04 04:37:29 +0000
@@ -315,13 +315,17 @@
         count = self._iterate_over(relpaths, gather, pb, 'stat', expand=False)
         return stats
 
+    def listable(self):
+        """Return True if this store supports listing."""
+        raise NotImplementedError
 
     def list_dir(self, relpath):
         """Return a list of all files at the given location.
         WARNING: many transports do not support this, so trying avoid using
         it if at all possible.
         """
-        raise NotImplementedError
+        raise TransportNotPossible("This transport has not "
+                                   "implemented list_dir.")
 
     def lock_read(self, relpath):
         """Lock the given file for shared (read) access.

=== modified file 'bzrlib/transport/http.py'
--- bzrlib/transport/http.py	2005-10-02 21:51:29 +0000
+++ bzrlib/transport/http.py	2005-10-04 04:37:29 +0000
@@ -216,12 +216,9 @@
         """Delete the item at relpath"""
         raise TransportNotPossible('http does not support delete()')
 
-    def list_dir(self, relpath):
-        """Return a list of all files at the given location.
-        WARNING: many transports do not support this, so trying avoid using
-        it if at all possible.
-        """
-        raise TransportNotPossible('http does not support list_dir()')
+    def listable(self):
+        """See Transport.listable."""
+        return False
 
     def stat(self, relpath):
         """Return the stat information for a file.

=== modified file 'bzrlib/transport/local.py'
--- bzrlib/transport/local.py	2005-10-02 21:51:29 +0000
+++ bzrlib/transport/local.py	2005-10-04 04:37:29 +0000
@@ -189,6 +189,9 @@
         else:
             return super(LocalTransport, self).copy_to(relpaths, other, pb=pb)
 
+    def listable(self):
+        """See Transport.listable."""
+        return True
 
     def list_dir(self, relpath):
         """Return a list of all files at the given location.

=== modified file 'bzrlib/upgrade.py'
--- bzrlib/upgrade.py	2005-10-05 09:08:22 +0000
+++ bzrlib/upgrade.py	2005-10-06 02:51:02 +0000
@@ -323,81 +323,56 @@
         parent_invs = map(self._load_updated_inventory, present_parents)
         for file_id in inv:
             ie = inv[file_id]
-            self._set_revision(rev, ie, parent_invs)
-            if not ie.has_text():
-                continue
             self._convert_file_version(rev, ie, parent_invs)
 
-
-    def _set_revision(self, rev, ie, parent_invs):
-        """Set name version for a file.
-        """
-        file_id = ie.file_id
-        if ie.kind == 'root_directory':
-            return
-        if len(parent_invs) == 0:
-            ie.revision = rev.revision_id
-        else:
-            matched_id = None
-            for p_inv in parent_invs:
-                if not p_inv.has_id(file_id):
-                    ie.revision = rev.revision_id
-                    break
-                old_ie = p_inv[file_id]
-                if (old_ie.parent_id != ie.parent_id
-                    or old_ie.name != ie.name
-                    or old_ie.text_sha1 != ie.text_sha1
-                    or old_ie.text_size != ie.text_size):
-                    ie.revision = rev.revision_id
-                    break
-                if matched_id is None:
-                    matched_id = old_ie.revision
-                elif matched_id != old_ie.revision:
-                    ie.revision = rev.revision_id
-                    break
-            else:
-                assert matched_id is not None
-                ie.revision = matched_id
-
     def _convert_file_version(self, rev, ie, parent_invs):
         """Convert one version of one file.
 
         The file needs to be added into the weave if it is a merge
         of >=2 parents or if it's changed from its parent.
         """
+        if ie.kind == 'root_directory':
+            return
         file_id = ie.file_id
         rev_id = rev.revision_id
         w = self.text_weaves.get(file_id)
         if w is None:
             w = Weave(file_id)
             self.text_weaves[file_id] = w
-        file_parents = []
         text_changed = False
-        for parent_inv in parent_invs:
-            if parent_inv.has_id(file_id):
-                parent_ie = parent_inv[file_id]
-                old_revision = parent_ie.revision
+        previous_entries = ie.find_previous_heads(parent_invs, w)
+        for old_revision in previous_entries:
                 # if this fails, its a ghost ?
                 assert old_revision in self.converted_revs 
-                if old_revision not in file_parents:
-                    file_parents.append(old_revision)
-                if parent_ie.text_sha1 != ie.text_sha1:
-                    text_changed = True
-        if len(file_parents) != 1 or text_changed:
+        self.snapshot_ie(previous_entries, ie, w, rev_id)
+        del ie.text_id
+        assert getattr(ie, 'revision', None) is not None
+
+    def snapshot_ie(self, previous_revisions, ie, w, rev_id):
+        # TODO: convert this logic, which is ~= snapshot to
+        # a call to:. This needs the path figured out. rather than a work_tree
+        # a v4 revision_tree can be given, or something that looks enough like
+        # one to give the file content to the entry if it needs it.
+        # and we need something that looks like a weave store for snapshot to 
+        # save against.
+        #ie.snapshot(rev, PATH, previous_revisions, REVISION_TREE, InMemoryWeaveStore(self.text_weaves))
+        if len(previous_revisions) == 1:
+            previous_ie = previous_revisions.values()[0]
+            if ie._unchanged(previous_ie):
+                ie.revision = previous_ie.revision
+                return
+        parent_indexes = map(w.lookup, previous_revisions)
+        if ie.has_text():
             file_lines = self.branch.text_store[ie.text_id].readlines()
             assert sha_strings(file_lines) == ie.text_sha1
             assert sum(map(len, file_lines)) == ie.text_size
-            w.add(rev_id, file_parents, file_lines, ie.text_sha1)
-            ie.revision = rev_id
+            w.add(rev_id, parent_indexes, file_lines, ie.text_sha1)
             self.text_count += 1
-            ##mutter('import text {%s} of {%s}',
-            ##       ie.text_id, file_id)
         else:
-            ##mutter('text of {%s} unchanged from parent', file_id)
-            ie.revision = file_parents[0]
-        del ie.text_id
-
-
+            w.add(rev_id, parent_indexes, [], None)
+        ie.revision = rev_id
+        ##mutter('import text {%s} of {%s}',
+        ##       ie.text_id, file_id)
 
     def _make_order(self):
         """Return a suitable order for importing revisions.

